# Copyright (c) 2009-2011,2013 Turbulenz Limited
"""
Turbulenz framework for creating browser based games.
"""

__author__ = "Turbulenz Limited"
__copyright__   = "Copyright (c) 2009-2011,2013 Turbulenz Limited"

__version__ = '1.0.1'
