# Copyright (c) 2012-2013 Turbulenz Limited

# Must be kept in sync with the version in the plugin and webgl engine
# code.  Python tools use this to stamp build products and to generate
# version checking code.
SDK_VERSION = "0.26.0.0"
